# Copyright 2022 Zurich Instruments AG
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import dataclasses
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from laboneq.compiler.code_generator.measurement_calculator import IntegrationTimes
from laboneq.compiler.common.device_type import DeviceType
from laboneq.compiler.experiment_access.experiment_dao import ExperimentDAO
from laboneq.core.exceptions import LabOneQException
from laboneq.core.types.enums.acquisition_type import is_spectroscopy
from laboneq.data.compilation_job import ParameterInfo
from laboneq.data.recipe import (
    AWG,
    IO,
    AcquireLength,
    Gains,
    Initialization,
    IntegratorAllocation,
    Measurement,
    NtStepKey,
    OscillatorParam,
    RealtimeExecutionInit,
    Recipe,
    RefClkType,
    SignalType,
    TriggeringMode,
)

if TYPE_CHECKING:
    from laboneq.compiler.workflow.compiler import LeaderProperties

_logger = logging.getLogger(__name__)


class RecipeGenerator:
    def __init__(self):
        self._recipe = Recipe()

    def add_oscillator_params(self, experiment_dao: ExperimentDAO):
        for signal_id in experiment_dao.signals():
            signal_info = experiment_dao.signal_info(signal_id)
            oscillator_info = experiment_dao.signal_oscillator(signal_id)
            if oscillator_info is None:
                continue
            if oscillator_info.is_hardware:
                if isinstance(oscillator_info.frequency, ParameterInfo):
                    frequency, param = None, oscillator_info.frequency.uid
                else:
                    frequency, param = oscillator_info.frequency, None

                for ch in signal_info.channels:
                    self._recipe.oscillator_params.append(
                        OscillatorParam(
                            id=oscillator_info.uid,
                            device_id=signal_info.device.uid,
                            channel=ch,
                            frequency=frequency,
                            param=param,
                        )
                    )

    def add_integrator_allocations(
        self,
        integration_unit_allocation,
        experiment_dao: ExperimentDAO,
        integration_weights,
    ):
        def _make_integrator_allocation(
            signal_id: str, integrator
        ) -> IntegratorAllocation:
            is_multistate = "signals" in integrator and len(integrator["signals"]) > 1
            signals = integrator.get("signals") or [signal_id]
            thresholds: list[float] = [
                experiment_dao.threshold(signal_id) for signal_id in signals
            ]
            weights: list[dict[str, Any]] = [
                next(iter(integration_weights[signal_id].values()), {})
                if signal_id in integration_weights
                else {}
                for signal_id in signals
            ]
            integrator_allocation = IntegratorAllocation(
                signal_id=signals if is_multistate else signal_id,
                device_id=integrator["device_id"],
                awg=integrator["awg_nr"],
                channels=integrator["channels"],
                weights=[w.get("basename") for w in weights]
                if is_multistate
                else weights[0].get("basename"),
            )
            if is_multistate or thresholds[0] is not None:
                integrator_allocation.threshold = (
                    thresholds if is_multistate else thresholds[0]
                )
            return integrator_allocation

        for signal_id, integrator in integration_unit_allocation.items():
            self._recipe.integrator_allocations.append(
                _make_integrator_allocation(signal_id, integrator)
            )

    def add_acquire_lengths(self, integration_times: IntegrationTimes):
        self._recipe.acquire_lengths.extend(
            [
                AcquireLength(
                    section_id=section_id,
                    signal_id=signal_id,
                    acquire_length=integration_info.length_in_samples,
                )
                for section_id, section_integration_time in integration_times.items()
                for signal_id, integration_info in section_integration_time.items()
                if not integration_info.is_play
            ]
        )

    def add_devices_from_experiment(self, experiment_dao: ExperimentDAO):
        for device in experiment_dao.device_infos():
            self._recipe.initializations.append(
                Initialization(
                    device_uid=device.uid, device_type=device.device_type.name
                )
            )

    def _find_initialization(self, device_uid) -> Initialization:
        for initialization in self._recipe.initializations:
            if initialization.device_uid == device_uid:
                return initialization
        raise LabOneQException(
            f"Internal error: missing initialization for device {device_uid}"
        )

    def add_connectivity_from_experiment(
        self,
        experiment_dao: ExperimentDAO,
        leader_properties: LeaderProperties,
        clock_settings: Dict[str, Any],
    ):
        if leader_properties.global_leader is not None:
            initialization = self._find_initialization(leader_properties.global_leader)
            initialization.config.repetitions = 1
            initialization.config.holdoff = 0
            if leader_properties.is_desktop_setup:
                initialization.config.triggering_mode = TriggeringMode.DESKTOP_LEADER
        if leader_properties.is_desktop_setup:
            # Internal followers are followers on the same device as the leader. This
            # is necessary for the standalone SHFQC, where the SHFSG part does neither
            # appear in the PQSC device connections nor the DIO connections.
            for f in leader_properties.internal_followers:
                initialization = self._find_initialization(f)
                initialization.config.triggering_mode = TriggeringMode.INTERNAL_FOLLOWER

        for device in experiment_dao.device_infos():
            device_uid = device.uid
            initialization = self._find_initialization(device_uid)
            reference_clock = experiment_dao.device_reference_clock(device_uid)
            initialization.config.reference_clock = (
                RefClkType._10MHZ if reference_clock == 10e6 else RefClkType._100MHZ
            )

            if (
                device.device_type.value == "hdawg"
                and clock_settings["use_2GHz_for_HDAWG"]
            ):
                initialization.config.sampling_rate = (
                    DeviceType.HDAWG.sampling_rate_2GHz
                )

            if device.device_type.value == "shfppc":
                ppchannels = []
                for signal in experiment_dao.signals():
                    amplifier_pump = experiment_dao.amplifier_pump(signal)
                    if (
                        amplifier_pump is None
                        or amplifier_pump.device.uid != device_uid
                    ):
                        continue
                    amplifier_pump_dict: dict[str, str | float | bool | int] = {
                        "cancellation": amplifier_pump.cancellation,
                        "alc_engaged": amplifier_pump.alc_engaged,
                        "use_probe": amplifier_pump.use_probe,
                        "channel": amplifier_pump.channel,
                    }
                    for field in [
                        "pump_freq",
                        "pump_power",
                        "probe_frequency",
                        "probe_power",
                    ]:
                        val = getattr(amplifier_pump, field)
                        if val is None:
                            continue
                        if isinstance(val, ParameterInfo):
                            amplifier_pump_dict[field] = val.uid
                        else:
                            amplifier_pump_dict[field] = val

                    ppchannels.append(amplifier_pump_dict)
                initialization.ppchannels = ppchannels

            for follower in experiment_dao.dio_followers():
                initialization = self._find_initialization(follower)
                if leader_properties.is_desktop_setup:
                    initialization.config.triggering_mode = (
                        TriggeringMode.DESKTOP_DIO_FOLLOWER
                    )
                else:
                    initialization.config.triggering_mode = TriggeringMode.DIO_FOLLOWER

        for pqsc_device_id in experiment_dao.pqscs():
            for port in experiment_dao.pqsc_ports(pqsc_device_id):
                follower_device_init = self._find_initialization(port["device"])
                follower_device_init.config.triggering_mode = (
                    TriggeringMode.ZSYNC_FOLLOWER
                )

    def add_output(
        self,
        device_id,
        channel,
        offset=0.0,
        diagonal=1.0,
        off_diagonal=0.0,
        precompensation=None,
        modulation=False,
        oscillator=None,
        oscillator_frequency=None,
        lo_frequency=None,
        port_mode=None,
        output_range=None,
        output_range_unit=None,
        port_delay=None,
        scheduler_port_delay=0.0,
        marker_mode=None,
        amplitude=None,
    ):
        if precompensation is not None:
            precomp_dict = {
                k: v
                for k, v in dataclasses.asdict(precompensation).items()
                if k in ("exponential", "high_pass", "bounce", "FIR")
            }
            if "clearing" in (precomp_dict["high_pass"] or {}):
                del precomp_dict["high_pass"]["clearing"]
        else:
            precomp_dict = None
        output = IO(
            channel=channel,
            enable=True,
            offset=offset,
            precompensation=precomp_dict,
            lo_frequency=lo_frequency,
            port_mode=port_mode,
            range=None if output_range is None else float(output_range),
            range_unit=output_range_unit,
            modulation=modulation,
            port_delay=port_delay,
            scheduler_port_delay=scheduler_port_delay,
            marker_mode=marker_mode,
            amplitude=amplitude,
        )
        if diagonal is not None and off_diagonal is not None:
            output.gains = Gains(diagonal=diagonal, off_diagonal=off_diagonal)

        initialization = self._find_initialization(device_id)
        initialization.outputs.append(output)

    def add_input(
        self,
        device_id,
        channel,
        lo_frequency=None,
        input_range=None,
        input_range_unit=None,
        port_delay=None,
        scheduler_port_delay=0.0,
    ):
        input = IO(
            channel=channel,
            enable=True,
            lo_frequency=lo_frequency,
            range=None if input_range is None else float(input_range),
            range_unit=input_range_unit,
            port_delay=port_delay,
            scheduler_port_delay=scheduler_port_delay,
        )

        initialization = self._find_initialization(device_id)
        initialization.inputs.append(input)

    def add_awg(
        self,
        device_id: str,
        awg_number: int,
        signal_type: str,
        qa_signal_id: Optional[str],
        command_table_match_offset: Optional[int],
        feedback_register: Optional[int],
    ):
        awg = AWG(
            awg=awg_number,
            signal_type=SignalType(signal_type),
            qa_signal_id=qa_signal_id,
            command_table_match_offset=command_table_match_offset,
            feedback_register=feedback_register,
        )
        initialization = self._find_initialization(device_id)
        initialization.awgs.append(awg)

    def add_realtime_step(
        self,
        device_id: str,
        awg_id: int,
        seqc_filename: str,
        wave_indices_name: str,
        nt_loop_indices: List[int],
    ):
        self._recipe.realtime_execution_init.append(
            RealtimeExecutionInit(
                device_id=device_id,
                awg_id=awg_id,
                seqc_ref=seqc_filename,
                wave_indices_ref=wave_indices_name,
                nt_step=NtStepKey(indices=tuple(nt_loop_indices)),
            )
        )

    def from_experiment(
        self,
        experiment_dao: ExperimentDAO,
        leader_properties: LeaderProperties,
        clock_settings: Dict[str, Any],
    ):
        self.add_devices_from_experiment(experiment_dao)
        self.add_connectivity_from_experiment(
            experiment_dao, leader_properties, clock_settings
        )
        self._recipe.is_spectroscopy = is_spectroscopy(experiment_dao.acquisition_type)

    def add_simultaneous_acquires(
        self, simultaneous_acquires: Dict[float, Dict[str, str]]
    ):
        # Keys are of no interest, only order and simultaneity is important
        self._recipe.simultaneous_acquires = list(simultaneous_acquires.values())

    def add_total_execution_time(self, total_execution_time):
        self._recipe.total_execution_time = total_execution_time

    def add_max_step_execution_time(self, max_step_execution_time):
        self._recipe.max_step_execution_time = max_step_execution_time

    def add_measurements(self, measurement_map: dict[str, list[dict]]):
        for initialization in self._recipe.initializations:
            device_uid = initialization.device_uid
            if device_uid in measurement_map:
                initialization.measurements = [
                    Measurement(
                        length=m.get("length"),
                        channel=m.get("channel"),
                    )
                    for m in measurement_map[device_uid]
                ]

    def recipe(self) -> Recipe:
        return self._recipe
